/*
 * Kerio IP filter
 * Copyright (C) Kerio Technologies s.r.o.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __KIPF_DEFS_H__
#define __KIPF_DEFS_H__

/* Version of this driver */
#define	KIPF_MAJOR_VERSION			2
#define	KIPF_MINOR_VERSION			3

/* Out 32 bit ptr type */
typedef	void* KIPF_PTR;

/*	Max count of buffers for request */
#define MAX_RW_BUFFERS_COUNT		64

/* rules for set_default_traffic_rule() */
#define TRAFFIC_BLOCKED				0
#define TRAFFIC_ALLOWED				1

/*	Maximum length of interface name */
#define KIPF_MAX_NSIZE				128

/*	Maximum length of kipf interface hardware address */
#define KIPF_HW_ADDR_LEN			6

/*	Max bytes to copy (incl. HW hdr.) to userspace in read() */
#define KIPF_MAX_BYTES_TO_COPY 128

#pragma pack(1)

/* 
 * common part of network packet and device packet
 */
struct kipf_header {
	/*	
	 * magic number 
	 *    - type of packet in kernel->userspace case 
	 *    - action in userspace->kernel case
	 */
	unsigned int magic;
#define	KIPF_MAGIC_FREE				0x0
#define	KIPF_MAGIC_IFACE			0x1
#define	KIPF_MAGIC_PACKET			0x2

#define	KIPF_ACTION_NULL			0x3
#define	KIPF_ACTION_ACCEPT			0x4
#define	KIPF_ACTION_DROP			0x5
#define	KIPF_ACTION_COMPLETE		0x6
#define	KIPF_ACTION_TX_IP_PACKET	0x7
#define	KIPF_ACTION_RX_ETH_FRAME	0x8
#define	KIPF_ACTION_TX_ETH_FRAME	0x9
};

struct kipf_iface {
	struct kipf_header header;			/* struct for packet recognizing */
	unsigned int ifindex;				/* interface index */
	char name[KIPF_MAX_NSIZE];			/* name of interface	*/
	char id[KIPF_MAX_NSIZE];			/* id of interface	*/
	char hw_addr[KIPF_HW_ADDR_LEN];		/* hardware address */
	unsigned int mtu;					/* Maximum transmission unit	*/

	unsigned int flags;
#define KIPF_IFACE_UP	0x1				/* interface is up		*/
#define KIPF_CARRIER_UP	0x2				/* carrier (cable status) */

	unsigned int hw_type;
#define KIPF_HW_TYPE_UNK	0x0			/* unknown	*/
#define KIPF_HW_TYPE_ETH	0x1			/* ethernet	*/
#define KIPF_HW_TYPE_PPP	0x2			/* ppp	*/
#define KIPF_HW_TYPE_VPN	0x3			/* vpn	*/
};

struct kipf_route_info {
	union {
		unsigned int next_hop;
		struct in6_addr next_hop6;
	};
	unsigned int ifindex;
};

struct kipf_ipsec_info {
	unsigned int type;
#define KIPF_IPSEC_TYPE_UNKNOWN		0x0
#define KIPF_IPSEC_TYPE_TRANSPORT	0x1
#define KIPF_IPSEC_TYPE_TUNNEL		0x2
	unsigned int reqId;
};

struct kipf_packet_k2u {
	unsigned int ifindex;				/*	ifindex of interface where packet was received/sent	*/
	unsigned int received;				/*	packet was received ? */
	unsigned int len;					/*	total length of received/sent packet with hw header. Not all the data are copied in read()! */
	struct kipf_ipsec_info ipsec;
};

struct kipf_packet_u2k {
	unsigned int modif_len;				/*	length of modified part of packet WTIHOUT hw header */
	struct kipf_route_info route_info;
};

struct kipf_packet {
	struct kipf_header header;			/*	struct for packet recognizing		*/
	void* id;					/*	packet id (skb*, but don't tell)    */

	//union { // union commented out, k2u would be damaged after complete()
		struct kipf_packet_k2u k2u;
		struct kipf_packet_u2k u2k;
	//};

	char pad[/*NET_IP_ALIGN*/2];

	char data[0];						/*	packet data  */
};

struct kipf_info {

	/* maximum size of buffer for reading/writing (currently hardcoded to MAX_RW_BUFFER_SIZE) including kipf_packet struct */
	unsigned int max_rw_buffer_size;

	struct {
		unsigned int major;
		unsigned int minor;
	} api_version;
};
typedef struct kipf_info KIPF_INFO;

struct kipf_request {

	/* input - count of pointers in header array */
	/* output - count of successfully read/handled buffers */
	unsigned int count;

	/* here is stored pointers to kipf_iface or kipf_packet depending on magic number */
	KIPF_PTR packets[MAX_RW_BUFFERS_COUNT];
};

struct ioctl_request {
	unsigned int action_type;

	union {
		struct {
			unsigned char *pkt;
			unsigned int pkt_len;
			struct kipf_route_info route_info;
		};

		KIPF_PTR packet;
	};
};
typedef struct ioctl_request IOCTL_REQ;

struct nf_iface_mode {
#define NF_IFACE_MODE_DISABLED	0x00
#define NF_IFACE_MODE_ENABLED	0x02
	unsigned int mode;
	unsigned int ifindex;				/*	ifindex of interface */
};
typedef struct nf_iface_mode NF_IFACE_MODE;

#pragma pack()

/*
** interface id - this code is shared with userspace
*/

typedef enum IfaceIdType {
	IIT_Eth,
	IIT_Vlan,
	IIT_Ppp,
	IIT_Bridge,
	IIT_Sit,
	IIT_Wlan
} IfaceIdType;

/* Bug 51001 - Create persistent ID of Kerio VPN adapter */
static const unsigned char KVNET_MAC[6] = {0x44, 0x45, 0x53, 0x54, 0x4f, 0x53};

static const struct { char *prefix; unsigned int shift; } ifaceIdParams[] = {
	{"eth", 0},
	{"vlan", 0},
	{"ppp", 3},
	{"br", 2},
	{"sit", 0},
	{"wlan", 0},
};

static inline char *name2IfaceId(IfaceIdType type, char *buff, int size, const char *name)
{
	unsigned int shift = ifaceIdParams[type].shift;

	if (shift && (strlen(name) <= shift || !isdigit((unsigned char)name[shift]))) {
		shift = 0;
	}
	snprintf(buff, size, "/device/%s/%s", ifaceIdParams[type].prefix, name + shift);
	buff[size - 1] = 0;
	return buff;
};

static inline char *mac2IfaceId(IfaceIdType type, char *buff, int size, unsigned char *hwAddress, int vlanId, const char *name)
{
	char vlan[32] = {0};
	const char * suffix = "";

	if (memcmp(hwAddress, KVNET_MAC, 6) == 0) {
		strcpy(buff, "KerioVPN"); // Bug 51001 - Create persistent ID of Kerio VPN adapter
		return buff;
	}

	if (type == IIT_Vlan) {
		snprintf(vlan, sizeof(vlan), "/%d", vlanId);
	}
	
	if (type == IIT_Wlan) {
		suffix = strchr(name, '_');
		if ( ! suffix) {
			suffix = "";
		}
	}
	
	snprintf(buff, size, "/device/%s/{%02x-%02x-%02x-%02x-%02x-%02x}%s%s", ifaceIdParams[type].prefix,
		hwAddress[0], hwAddress[1], hwAddress[2], hwAddress[3], hwAddress[4], hwAddress[5], suffix, vlan);

	buff[size - 1] = 0;
	return buff;
};

static inline int isWlanIfaceId(const char *ifaceId) 
{
	return strncmp(ifaceId,"/device/wlan/", 13) == 0 ? 1 : 0 ;
}

#endif // __KIPF_DEFS_H__
