/*
 * Kerio IP filter
 * Copyright (C) Kerio Technologies s.r.o.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "hdrs.h"
#include "io_read.h"
#include "io_write.h"

#define MAX_COUNT_OF_DEVICES 256

struct io_priv {
	pid_t pid;
	int dummy_dev;
	int snort;
};

int gs_canceled = 0;
int gs_major = -1;
int gs_minor = -1;
int gs_io_dev_added = 0;
struct cdev gs_cdev;
struct class *gs_ipf_class = NULL;
struct device *gs_device = NULL;

DECLARE_WAIT_QUEUE_HEAD(read_wait);

int g_device_opened = 0;

void
io_dev_wakeup(
	void
	)
{
	//TRACE_DBG("%s: wake up pending io request!\n", __func__);

	wake_up_interruptible(&read_wait);
}

static inline ssize_t
io_dev_iov_total(
	const struct iovec *iv,
	unsigned long count
	)
{
	unsigned long i = 0;
	ssize_t len = 0;

	for (i = 0, len = 0; i < count; i++) {
		len += iv[i].iov_len;
	}

	return len;
}

static ssize_t
io_dev_aio_read(
	struct kiocb *iocb,
	const struct iovec *iv,
	unsigned long count,
	loff_t pos
	)
{
	struct file *filp = iocb->ki_filp;
	ssize_t len = 0;

	/* if we acts as dummy device simple return 0 as we are not able to read anything */
	if (((struct io_priv*)filp->private_data)->dummy_dev) {
		return len;
	}

	if ((len = io_dev_iov_total(iv, count)) < 0) {
		TRACE(LOG_WARN, "%s: io_dev_iov_total(): len %d", __func__, len);
		return -EINVAL;
	}

	return io_read(iocb->ki_filp, iv, len);
}

static ssize_t
io_dev_aio_write(
	struct kiocb *iocb,
	const struct iovec *iv,
	unsigned long count,
	loff_t pos
	)
{
	struct file *filp = iocb->ki_filp;
	ssize_t len = 0;

	if ((len = io_dev_iov_total(iv, count)) < 0) {
		TRACE(LOG_WARN, "%s: io_dev_iov_total(): len %d", __func__, len);
		return -EINVAL;
	}

	/* if we acts as dummy device simple return count of bytes as we write all */
	if (((struct io_priv*)filp->private_data)->dummy_dev) {
		return len;
	}

	return io_write(iocb->ki_filp, iv, len);
}

/**
 *	io_dev_poll (select)
 *
 *	POLLIN      Data is available on the file descriptor for reading.
 *
 *	POLLNORM    Same as POLLIN.
 *
 *  POLLPRI     Same as POLLIN.
 *
 *  POLLOUT     Data can be written to the file descriptor without blocking.
 *
 *  POLLERR     This flag is not used in this implementation and is provided
 *              only for source code compatibility.
 *
 *  POLLHUP     The file descriptor was valid before the polling process and
 *              invalid after.  Presumably, this means that the file descrip-
 *              tor was closed sometime during the poll.
 *
 *  POLLNVAL    The corresponding file descriptor is invalid.
 *
 *  POLLRDNORM  Same as POLLIN.
 *
 *  POLLRDBAND  Same as POLLIN.
 *
 *  POLLWRNORM  Same as POLLOUT.
 *
 *  POLLWRBAND  Same as POLLOUT.
 *
 *  POLLMSG     This flag is not used in this implementation and is provided
 *              only for source code compatibility.
 *
 */
static unsigned int
io_dev_poll(
	struct file *filp,
	poll_table *wait
	)
{
	unsigned int mask = POLLPRI | POLLERR;
	int pkts = 0;

	/* if we acts as dummy device simple return */
	if (((struct io_priv*)filp->private_data)->dummy_dev) {
		return mask;
	}

	if (unlikely(gs_canceled)) {
		TRACE(LOG_INFO, "%s: io request has been canceled.", __func__);
		return mask;
	}

	do {

		/* only insert into poll table and immediately return */
		poll_wait(filp, &read_wait, wait);

		/* we are always writable */
		mask = POLLOUT | POLLWRNORM;

		/* some packets in main queue? */
		if ((pkts = queue_len(&g_main_queue))) {
			mask |= POLLIN | POLLRDNORM;
		}

		TRACE_DBG("%s: main queue contains %d packet(s).\n", __func__, pkts);

	} while(0);

	return mask;
}

static int
ioctl_set_nf_iface_mode(
	unsigned long arg
	)
{
	NF_IFACE_MODE if_mode;
	struct net_device *dev = NULL;
	struct net_device_extension *ext = NULL;

	if (copy_from_user(&if_mode, (NF_IFACE_MODE __user *) arg, sizeof(NF_IFACE_MODE))) {
		TRACE(LOG_WARN, "%s: get_user(): segmentation fault.", __func__);
		return -EFAULT;
	}

	if (!(dev = netdev_find_by_ifindex(if_mode.ifindex))) {
		TRACE(LOG_WARN, "%s: netdev_find_by_ifindex(): can't find net device with ifindex %d", 	__func__, if_mode.ifindex);
		return -EINVAL;
	}

	if (!(ext = netdev_get_extension(dev))) {
		TRACE(LOG_WARN, "%s: netdev_get_extension(): can't find device extension on '%s'.'", __func__, dev->name);
		return -EINVAL;
	}

	/* set new mode */
	ext->mode = if_mode.mode;

	TRACE_DBG("%s: NF mode on iface %d has been set to: %s\n", __func__, if_mode.ifindex,
		((if_mode.mode == NF_IFACE_MODE_ENABLED) ? ("enabled") : ("disabled")));

	return 0;
}

static int
ioctl_get_info(
	unsigned long arg
	)
{
	struct kipf_info *info = (struct kipf_info __user *) arg;

	if (put_user(g_linipf_cfg.max_packet_size, &info->max_rw_buffer_size)) {
		TRACE(LOG_WARN, "%s: put_user(): segmentation fault.", __func__);
		return -EFAULT;
	}

	if (put_user(KIPF_MAJOR_VERSION, &info->api_version.major)) {
		TRACE(LOG_WARN, "%s: put_user(): segmentation fault.", __func__);
		return -EFAULT;
	}

	if (put_user(KIPF_MINOR_VERSION, &info->api_version.minor)) {
		TRACE(LOG_WARN, "%s: put_user(): segmentation fault.", __func__);
		return -EFAULT;
	}

	TRACE(LOG_INFO, "%s: API version: %d.%d, max buffer size: %d",
		__func__, KIPF_MAJOR_VERSION, KIPF_MINOR_VERSION, g_linipf_cfg.max_packet_size);

	return 0;
}

static int
ioctl_set_gso_config(
	unsigned long arg
	)
{
	unsigned int enable = 0;

	if (get_user(enable, (unsigned int __user *) arg)) {
		TRACE(LOG_WARN, "%s: get_user(): segmentation fault.", __func__);
		return -EFAULT;
	}

	TRACE(LOG_INFO, "%s: Offloading has been %s.", __func__, enable ? "enabled" : "disabled");
	g_linipf_cfg.gso_enabled = enable;

	return 0;
}

static int
ioctl_set_default_traffic_rule(
	unsigned long arg
	)
{
	unsigned int rule = 0;

	if (get_user(rule, (unsigned int __user *) arg)) {
		TRACE(LOG_WARN, "%s: get_user(): segmentation fault.", __func__);
		return -EFAULT;
	}

	g_linipf_cfg.block_traffic_when_api_is_closed = ((rule == TRAFFIC_BLOCKED) ? (1) : (0));

	if (g_linipf_cfg.block_traffic_when_api_is_closed) {
		g_linipf_cfg.start_time = 0xffffffff - START_TIMEOUT;
		g_linipf_cfg.timeout_expired = 0;
	} else {
		g_linipf_cfg.start_time = 0;
		g_linipf_cfg.timeout_expired = 1;
	}

	TRACE(LOG_INFO, "%s: default traffic rule was set to: %s",
		__func__, ((g_linipf_cfg.block_traffic_when_api_is_closed) ? ("drop") : ("accept")));

	return 0;
}

static int
ioctl_cancel_pending_read(
	unsigned long arg
	)
{
	TRACE(LOG_INFO, "%s: device mode set to canceled.", __func__);
	gs_canceled = 1;
	io_dev_wakeup();
	return 0;
}

static int
ioctl_handle_packet(
	unsigned long arg
	)
{
	IOCTL_REQ ioctlRequest;

	if (copy_from_user(&ioctlRequest, (IOCTL_REQ __user *) arg, sizeof(IOCTL_REQ))) {
		TRACE(LOG_WARN, "%s: get_user(): segmentation fault.", __func__);
		return -EFAULT;
	}
	
	switch (ioctlRequest.action_type) {
		case KIPF_ACTION_TX_IP_PACKET:
			return pkt_tx_ip(ioctlRequest.pkt, ioctlRequest.pkt_len, &ioctlRequest.route_info);

		case KIPF_ACTION_RX_ETH_FRAME:
		case KIPF_ACTION_TX_ETH_FRAME:
			return pkt_tx_rx_eth(ioctlRequest.pkt, ioctlRequest.pkt_len, (ioctlRequest.action_type == KIPF_ACTION_TX_ETH_FRAME), &ioctlRequest.route_info);
	}

	return io_write_handle_packet((struct kipf_packet*)ioctlRequest.packet);
}

long
io_dev_ioctl(
	struct file *filp,
	unsigned int cmd,
	unsigned long arg
	)
{
	struct io_priv *priv = filp->private_data;
	/* if we acts as dummy device simple return I/O error */
	if (cmd != KIPF_IOCTL_WR_STARTED && cmd != KIPF_IOCTL_SNORT_STARTED && priv->dummy_dev) {
		return -EIO;
	}
	switch (cmd) {
		case KIPF_IOCTL_SET_NF_IFACE_MODE:
			return ioctl_set_nf_iface_mode(arg);

		case KIPF_IOCTL_GET_INFO:
			return ioctl_get_info(arg);

		case KIPF_IOCTL_SET_DEFAULT_TRAFFIC_RULE:
			return ioctl_set_default_traffic_rule(arg);

		case KIPF_IOCTL_CANCEL_PENDING_READ:
			return ioctl_cancel_pending_read(arg);

		case KIPF_IOCTL_HANDLE_PACKET:
			return ioctl_handle_packet(arg);

		case KIPF_IOCTL_SNORT_STARTED:
			TRACE(LOG_INFO, "%s: '%s' recognized as non-privileged application.", __func__, current->comm);
			priv->snort = 1;
			ip_filter_snort_started(priv->pid);
			return 0;

		case KIPF_IOCTL_WR_STARTED:
			TRACE(LOG_INFO, "%s: '%s' recognized as privileged application.", __func__, current->comm);
			priv->dummy_dev = 0;
			/* kipf device already opened ? */
			if (g_device_opened) {
				kfree(priv);
				TRACE(LOG_ERR, "%s: device '/dev/kipf' already opened.", __func__);
				return -EPERM;
			} else {
				/* Enable packet queueing */
				queue_enable(&g_main_queue, 1);
				queue_enable(&g_pending_queue, 1);
				/* Signal all up net devices */
				netdev_signal_all_up();
				/* not cancelled yet */
				gs_canceled = 0;
				/* Enable ip_filter_hook() queueing */
				g_device_opened = 1;
			}
			return 0;

		case KIPF_IOCTL_GSO:
			return ioctl_set_gso_config(arg);

		default:
			TRACE(LOG_WARN, "%s: unimplemented ioctl command %08X", __func__, cmd);
			break;
	}

	return -EINVAL;
}

static int
io_dev_close(
	struct inode *inode,
	struct file *filp
	)
{
	struct io_priv *priv = filp->private_data;

	/* invalidate private data pointer */
	filp->private_data = NULL;

	if (priv->snort) {
		ip_filter_snort_stopped(priv->pid);
	}

	if (!priv->dummy_dev && g_device_opened) {
		struct sk_buff *skb = NULL;

		//
		//	Disable ip_filter_hook() queueing
		//
		g_device_opened = 0;

		//
		//	Disable packet queueing
		//
		queue_enable(&g_main_queue, 0);
		queue_enable(&g_pending_queue, 0);

		//
		//	Flush pending skb's
		//
		while ((skb = queue_remove(&g_main_queue, NULL, NULL))) {
			pkt_verdict(skb, NF_DROP);
		}

		while ((skb = queue_remove(&g_pending_queue, NULL, NULL))) {
			pkt_verdict(skb, NF_DROP);
		}

		if (IS_PTR_VALID(filp->private_data)) {
			filp->private_data = NULL;
		}
	}

	TRACE(LOG_INFO, "%s: device '/dev/kipf' closed by '%s', pid %d", __func__, current->comm, priv->pid);

	/* free memory for private data */
	kfree(priv);

	return 0;
}

static int
io_dev_open(
	struct inode *inode,
	struct file *filp
	)
{
	/* allocate memory for private data */
	struct io_priv *priv = kmalloc(sizeof(struct io_priv), GFP_KERNEL);
	if (!priv) {
		TRACE(LOG_ERR, "%s: alloc_nonpaged_mem(): can't allocate memory for private data.", __func__);
		return -ENOMEM;
	}
	memset(priv, 0, sizeof(struct io_priv));
	/* save pid */
	priv->pid = current->pid;
	/* dummy device by default, will be override by ioctl */
	priv->dummy_dev = 1;

	TRACE(LOG_INFO, "%s: device '/dev/kipf' opened by '%s', pid %d", __func__, current->comm, priv->pid);

	/* save private data */
	filp->private_data = priv;
	return 0;
}

static const struct file_operations kipf_fops = {
	.owner		= THIS_MODULE,
	.poll		= io_dev_poll,
	.unlocked_ioctl	= io_dev_ioctl,
	.open		= io_dev_open,
	.release	= io_dev_close,
	.read		= do_sync_read,
	.aio_read	= io_dev_aio_read,
	.write		= do_sync_write,
	.aio_write	= io_dev_aio_write
};

int
io_dev_init(
	void
	)
{
	dev_t dev;
	int err = 0;

	do {

		gs_ipf_class = ERR_PTR(-ENOMEM);
		gs_device = ERR_PTR(-ENOMEM);

		if((err = alloc_chrdev_region(&dev, 0, MAX_COUNT_OF_DEVICES, "kipf"))) {
			TRACE(LOG_ERR, "%s: alloc_chrdev_region(): %d", __func__, err);
			break;
		}

		gs_major = MAJOR(dev);
		gs_minor = MINOR(dev);

		cdev_init(&gs_cdev, &kipf_fops);

		gs_cdev.owner = THIS_MODULE;

		if ((err = cdev_add(&gs_cdev, dev, MAX_COUNT_OF_DEVICES))) {
			TRACE(LOG_ERR, "%s: cdev_add(): %d", __func__, err);
			break;
		}

		gs_io_dev_added = 1;

		if (IS_ERR((gs_ipf_class = class_create(THIS_MODULE, "kipf")))) {
			err = PTR_ERR(gs_ipf_class);
			TRACE(LOG_ERR, "%s: class_create(): %d", __func__, err);
			break;
		}

		if (IS_ERR((gs_device = device_create(gs_ipf_class, NULL, dev, NULL, "kipf")))) {
			err = PTR_ERR(gs_device);
			TRACE(LOG_ERR, "%s: device_create(): %d", __func__, err);
			break;
		}

		TRACE(LOG_INFO, "%s: device '/dev/kipf' created.", __func__);

	} while(0);

	return err;
}

int
io_dev_remove(
	void
	)
{
	if (gs_major != -1) {

		if (gs_io_dev_added) {

			if (!IS_ERR(gs_ipf_class)) {

				if (!IS_ERR(gs_device)) {
					device_destroy(gs_ipf_class, MKDEV(gs_major, gs_minor));
					gs_device = ERR_PTR(-ENOMEM);
				}

				class_destroy(gs_ipf_class);
				gs_ipf_class = ERR_PTR(-ENOMEM);
			}

			cdev_del(&gs_cdev);
			gs_io_dev_added = 0;
		}

		unregister_chrdev_region(MKDEV(gs_major, gs_minor), MAX_COUNT_OF_DEVICES);

		gs_major = -1;
		gs_minor = -1;

		TRACE(LOG_INFO, "%s: device '/dev/kipf' removed.", __func__);
	}

	return 0;
}
